@@ -11,18 +11,26 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import datetime
+import functools
+
 from six.moves.urllib import parse as urlparse
 
 from django.conf import settings
 from django.contrib import auth
-from django.contrib.auth.models import AnonymousUser
 from django.contrib.auth import middleware
+from django.contrib.auth import models
+from django.utils import decorators
 from django.utils import timezone
 
 from keystoneclient.v2_0 import client as client_v2
 from keystoneclient.v3 import client as client_v3
 
 
+_PROJECT_CACHE = {}
+
+_TOKEN_TIMEOUT_MARGIN = getattr(settings, 'TOKEN_TIMEOUT_MARGIN', 0)
+
 """
 We need the request object to get the user, so we'll slightly modify the
 existing django.contrib.auth.get_user method. To do so we update the
@@ -45,9 +53,9 @@ def get_user(request):
         backend_path = request.session[auth.BACKEND_SESSION_KEY]
         backend = auth.load_backend(backend_path)
         backend.request = request
-        user = backend.get_user(user_id) or AnonymousUser()
+        user = backend.get_user(user_id) or models.AnonymousUser()
     except KeyError:
-        user = AnonymousUser()
+        user = models.AnonymousUser()
     return user
 
 
@@ -59,77 +67,45 @@ def patch_middleware_get_user():
 """ End Monkey-Patching. """
 
 
-def check_token_expiration(token):
-    """ Timezone-aware checking of the auth token's expiration timestamp.
+def is_token_valid(token, margin=None):
+    """Timezone-aware checking of the auth token's expiration timestamp.
 
     Returns ``True`` if the token has not yet expired, otherwise ``False``.
+
+    .. param:: token
+
+       The openstack_auth.user.Token instance to check
+
+    .. param:: margin
+
+       A time margin in seconds to subtract from the real token's validity.
+       An example usage is that the token can be valid once the middleware
+       passed, and invalid (timed-out) during a view rendering and this
+       generates authorization errors during the view rendering.
+       A default margin can be set by the TOKEN_TIMEOUT_MARGIN in the
+       django settings.
     """
     expiration = token.expires
+    # In case we get an unparseable expiration timestamp, return False
+    # so you can't have a "forever" token just by breaking the expires param.
+    if expiration is None:
+        return False
+    if margin is None:
+        margin = getattr(settings, 'TOKEN_TIMEOUT_MARGIN', 0)
+    expiration = expiration - datetime.timedelta(seconds=margin)
     if settings.USE_TZ and timezone.is_naive(expiration):
         # Presumes that the Keystone is using UTC.
         expiration = timezone.make_aware(expiration, timezone.utc)
-    # In case we get an unparseable expiration timestamp, return False
-    # so you can't have a "forever" token just by breaking the expires param.
-    if expiration:
-        return expiration > timezone.now()
-    else:
-        return False
-
-
-# Copied from Keystone's keystone/common/cms.py file.
-PKI_ANS1_PREFIX = 'MII'
-
-
-def is_ans1_token(token):
-    '''
-    thx to ayoung for sorting this out.
-
-    base64 decoded hex representation of MII is 3082
-    In [3]: binascii.hexlify(base64.b64decode('MII='))
-    Out[3]: '3082'
-
-    re: http://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
-
-    pg4:  For tags from 0 to 30 the first octet is the identfier
-    pg10: Hex 30 means sequence, followed by the length of that sequence.
-    pg5:  Second octet is the length octet
-          first bit indicates short or long form, next 7 bits encode the number
-          of subsequent octets that make up the content length octets as an
-          unsigned binary int
-
-          82 = 10000010 (first bit indicates long form)
-          0000010 = 2 octets of content length
-          so read the next 2 octets to get the length of the content.
-
-    In the case of a very large content length there could be a requirement to
-    have more than 2 octets to designate the content length, therefore
-    requiring us to check for MIM, MIQ, etc.
-    In [4]: base64.b64encode(binascii.a2b_hex('3083'))
-    Out[4]: 'MIM='
-    In [5]: base64.b64encode(binascii.a2b_hex('3084'))
-    Out[5]: 'MIQ='
-    Checking for MI would become invalid at 16 octets of content length
-    10010000 = 90
-    In [6]: base64.b64encode(binascii.a2b_hex('3090'))
-    Out[6]: 'MJA='
-    Checking for just M is insufficient
-
-    But we will only check for MII:
-    Max length of the content using 2 octets is 7FFF or 32767
-    It's not practical to support a token of this length or greater in http
-    therefore, we will check for MII only and ignore the case of larger tokens
-    '''
-    return token[:3] == PKI_ANS1_PREFIX
+    return expiration > timezone.now()
 
 
 # From django.contrib.auth.views
 # Added in Django 1.4.3, 1.5b2
 # Vendored here for compatibility with old Django versions.
 def is_safe_url(url, host=None):
-    """
-    Return ``True`` if the url is a safe redirection (i.e. it doesn't point to
-    a different host).
+    """Return ``True`` if the url is a safe redirection.
 
+    The safe redirection means that it doesn't point to a different host.
     Always returns ``False`` on an empty url.
     """
     if not url:
@@ -138,6 +114,37 @@ def is_safe_url(url, host=None):
     return not netloc or netloc == host
 
 
+def memoize_by_keyword_arg(cache, kw_keys):
+    """Memoize a function using the list of keyword argument name as its key.
+
+    Wrap a function so that results for any keyword argument tuple are stored
+    in 'cache'. Note that the keyword args to the function must be usable as
+    dictionary keys.
+
+    :param cache: Dictionary object to store the results.
+    :param kw_keys: List of keyword arguments names. The values are used
+                    for generating the key in the cache.
+    """
+    def _decorator(func):
+        @functools.wraps(func, assigned=decorators.available_attrs(func))
+        def wrapper(*args, **kwargs):
+            mem_args = [kwargs[key] for key in kw_keys if key in kwargs]
+            mem_args = '__'.join(str(mem_arg) for mem_arg in mem_args)
+            if not mem_args:
+                return func(*args, **kwargs)
+            if mem_args in cache:
+                return cache[mem_args]
+            result = func(*args, **kwargs)
+            cache[mem_args] = result
+            return result
+        return wrapper
+    return _decorator
+
+
+def remove_project_cache(token):
+    _PROJECT_CACHE.pop(token, None)
+
+
 # Helper for figuring out keystone version
 # Implementation will change when API version discovery is available
 def get_keystone_version():
@@ -151,14 +158,38 @@ def get_keystone_client():
         return client_v3
 
 
+def has_in_url_path(url, sub):
+    """Test if the `sub` string is in the `url` path."""
+    scheme, netloc, path, query, fragment = urlparse.urlsplit(url)
+    return sub in path
+
+
+def url_path_replace(url, old, new, count=None):
+    """Return a copy of url with replaced path.
+
+    Return a copy of url with all occurrences of old replaced by new in the url
+    path.  If the optional argument count is given, only the first count
+    occurrences are replaced.
+    """
+    args = []
+    scheme, netloc, path, query, fragment = urlparse.urlsplit(url)
+    if count is not None:
+        args.append(count)
+    return urlparse.urlunsplit((
+        scheme, netloc, path.replace(old, new, *args), query, fragment))
+
+
+@memoize_by_keyword_arg(_PROJECT_CACHE, ('token', ))
 def get_project_list(*args, **kwargs):
     if get_keystone_version() < 3:
-        auth_url = kwargs.get('auth_url', '').replace('v3', 'v2.0')
+        auth_url = url_path_replace(
+            kwargs.get('auth_url', ''), '/v3', '/v2.0', 1)
         kwargs['auth_url'] = auth_url
         client = get_keystone_client().Client(*args, **kwargs)
         projects = client.tenants.list()
     else:
-        auth_url = kwargs.get('auth_url', '').replace('v2.0', 'v3')
+        auth_url = url_path_replace(
+            kwargs.get('auth_url', ''), '/v2.0', '/v3', 1)
         kwargs['auth_url'] = auth_url
         client = get_keystone_client().Client(*args, **kwargs)
         client.management_url = auth_url
